#
# inheritancegraph.rb
#
# Copyright (c) 2003 Minero Aoki <aamine@loveruby.net>
#
# This program is free software.
# You can distribute/modify this program under the terms of
# the GNU Lesser General Public License version 2 or later.
#

module ReFe

  class InvalidGraph < StandardError; end

  class InheritanceGraph

    def InheritanceGraph.parse( path )
      graph = new()
      File.foreach(path) do |line|
        next if line.sub(/\A\#.*/, '').strip.empty?
        dest, op, src = line.split
        case op
        when '<-'   # include
          graph.add_include_edge dest, src
        when '<'   # inheritance
          graph.add_class_node dest, src
        else
          raise "ReFe igraph: cannot parse: #{line.strip}"
        end
      end
      graph
    end

    def initialize
      @nodes = {}
      @nodes['Object'] = ClassNode.new('Object', nil)
    end

    def true_ancestors_of( name )
      ancestors_of(name)[1..-1]
    end

    def ancestors_of( name )
      return [name] unless @nodes.key?(name)
      @nodes[name].ancestors.map {|node| node.name }
    end

    def superclasses_of( name )
      return [] unless @nodes.key?(name)
      list = @nodes[name].ancestors.select {|n| n.class? }.map {|n| n.name }
      list.shift if list[0] == name
      list
    end

    def add_include_edge( dest, incl )
      raise InvalidGraph, "destination not exist: #{dest}" unless @nodes[dest]
      @nodes[dest].include module_node(incl)
    end

    def add_class_node( name, super_class )
      raise InvalidGraph, "same class defined twice: #{name}" if @nodes[name]
      @nodes[name] = ClassNode.new(name, class_node(super_class))
    end

    private

    def module_node( name )
      raise InvalidGraph, "not a module: #{name}" if @nodes[name] and @nodes[name].class?
      @nodes[name] ||= ModuleNode.new(name)
    end

    def class_node( name )
      @nodes[name] or raise InvalidGraph, "class not exist: #{name}"
    end
  
  end


  class ModuleNode

    def initialize( name )
      @name = name
      @included = []
    end

    attr_reader :name

    def inspect
      "\#<#{self.class} #{@name}>"
    end

    def class?
      false
    end

    def include( mod )
      @included.unshift mod
    end

    def ancestors
      result = [self]
      @included.each do |mod|
        result.concat mod.ancestors
      end
      result.uniq   # .uniq is ReFe specific
    end

  end


  class ClassNode < ModuleNode

    def initialize( name, super_node )
      super name
      @superclass = super_node
    end

    def class?
      true
    end

    def ancestors
      (super() + (@superclass ? @superclass.ancestors : [])).uniq
    end

  end

end
